@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007 Free Software Foundation
+ * Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008 Free Software Foundation
  *
- * Author: Nikos Mavroyanopoulos
+ * Author: Nikos Mavrogiannopoulos
  *
  * This file is part of GNUTLS.
  *
@@ -52,6 +52,7 @@
 #include <auth_anon.h>		/* for gnutls_anon_server_credentials_t */
 #include <auth_psk.h>		/* for gnutls_psk_server_credentials_t */
 #include <gc.h>
+#include <random.h>
 
 #ifdef HANDSHAKE_DEBUG
 #define ERR(x, y) _gnutls_handshake_log( "HSK[%x]: %s (%d)\n", session, x,y)
@@ -68,13 +69,12 @@ int _gnutls_server_select_comp_method (gnutls_session_t session,
 
 /* Clears the handshake hash buffers and handles.
  */
-inline static void
+static void
 _gnutls_handshake_hash_buffers_clear (gnutls_session_t session)
 {
-  _gnutls_hash_deinit (session->internals.handshake_mac_handle_md5, NULL);
-  _gnutls_hash_deinit (session->internals.handshake_mac_handle_sha, NULL);
-  session->internals.handshake_mac_handle_md5 = NULL;
-  session->internals.handshake_mac_handle_sha = NULL;
+  _gnutls_hash_deinit (&session->internals.handshake_mac_handle_md5, NULL);
+  _gnutls_hash_deinit (&session->internals.handshake_mac_handle_sha, NULL);
+  session->internals.handshake_mac_handle_init = 0;
   _gnutls_handshake_buffer_clear (session);
 }
 
@@ -146,23 +146,24 @@ static int
 _gnutls_ssl3_finished (gnutls_session_t session, int type, opaque * ret)
 {
   const int siz = SSL_MSG_LEN;
-  mac_hd_t td_md5;
-  mac_hd_t td_sha;
+  digest_hd_st td_md5;
+  digest_hd_st td_sha;
   const char *mesg;
+  int rc;
 
-  td_md5 = _gnutls_hash_copy (session->internals.handshake_mac_handle_md5);
-  if (td_md5 == NULL)
+  rc = _gnutls_hash_copy (&td_md5, &session->internals.handshake_mac_handle_md5);
+  if (rc < 0)
     {
       gnutls_assert ();
-      return GNUTLS_E_HASH_FAILED;
+      return rc;
     }
 
-  td_sha = _gnutls_hash_copy (session->internals.handshake_mac_handle_sha);
-  if (td_sha == NULL)
+  rc = _gnutls_hash_copy (&td_sha, &session->internals.handshake_mac_handle_sha);
+  if (rc < 0)
     {
       gnutls_assert ();
-      _gnutls_hash_deinit (td_md5, NULL);
-      return GNUTLS_E_HASH_FAILED;
+      _gnutls_hash_deinit (&td_md5, NULL);
+      return rc;
     }
 
   if (type == GNUTLS_SERVER)
@@ -174,13 +175,13 @@ _gnutls_ssl3_finished (gnutls_session_t session, int type, opaque * ret)
       mesg = SSL3_CLIENT_MSG;
     }
 
-  _gnutls_hash (td_md5, mesg, siz);
-  _gnutls_hash (td_sha, mesg, siz);
+  _gnutls_hash (&td_md5, mesg, siz);
+  _gnutls_hash (&td_sha, mesg, siz);
 
-  _gnutls_mac_deinit_ssl3_handshake (td_md5, ret,
+  _gnutls_mac_deinit_ssl3_handshake (&td_md5, ret,
 				     session->security_parameters.
 				     master_secret, TLS_MASTER_SIZE);
-  _gnutls_mac_deinit_ssl3_handshake (td_sha, &ret[16],
+  _gnutls_mac_deinit_ssl3_handshake (&td_sha, &ret[16],
 				     session->security_parameters.
 				     master_secret, TLS_MASTER_SIZE);
 
@@ -199,37 +200,38 @@ _gnutls_finished (gnutls_session_t session, int type, void *ret)
   opaque concat[36];
   size_t len;
   const char *mesg;
-  mac_hd_t td_md5 = NULL;
-  mac_hd_t td_sha;
+  digest_hd_st td_md5;
+  digest_hd_st td_sha;
   gnutls_protocol_t ver = gnutls_protocol_get_version (session);
+  int rc;
 
   if (ver < GNUTLS_TLS1_2)
     {
-      td_md5 = _gnutls_hash_copy (session->internals.handshake_mac_handle_md5);
-      if (td_md5 == NULL)
+      rc = _gnutls_hash_copy (&td_md5, &session->internals.handshake_mac_handle_md5);
+      if (rc < 0)
 	{
 	  gnutls_assert ();
-	  return GNUTLS_E_HASH_FAILED;
+	  return rc;
 	}
     }
 
-  td_sha = _gnutls_hash_copy (session->internals.handshake_mac_handle_sha);
-  if (td_sha == NULL)
+  rc = _gnutls_hash_copy (&td_sha, &session->internals.handshake_mac_handle_sha);
+  if (rc < 0)
     {
       gnutls_assert ();
-      _gnutls_hash_deinit (td_md5, NULL);
-      return GNUTLS_E_HASH_FAILED;
+      _gnutls_hash_deinit (&td_md5, NULL);
+      return rc;
     }
 
   if (ver < GNUTLS_TLS1_2)
     {
-      _gnutls_hash_deinit (td_md5, concat);
-      _gnutls_hash_deinit (td_sha, &concat[16]);
+      _gnutls_hash_deinit (&td_md5, concat);
+      _gnutls_hash_deinit (&td_sha, &concat[16]);
       len = 20 + 16;
     }
   else
     {
-      _gnutls_hash_deinit (td_sha, concat);
+      _gnutls_hash_deinit (&td_sha, concat);
       len = 20;
     }
 
@@ -253,6 +255,7 @@ int
 _gnutls_tls_create_random (opaque * dst)
 {
   uint32_t tim;
+  int ret;
 
   /* Use weak random numbers for the most of the
    * buffer except for the first 4 that are the
@@ -263,15 +266,70 @@ _gnutls_tls_create_random (opaque * dst)
   /* generate server random value */
   _gnutls_write_uint32 (tim, dst);
 
-  if (gc_nonce (&dst[4], TLS_RANDOM_SIZE - 4) != GC_OK)
+  ret = _gnutls_rnd (RND_NONCE, &dst[4], TLS_RANDOM_SIZE - 4);
+  if (ret < 0)
     {
       gnutls_assert ();
-      return GNUTLS_E_RANDOM_FAILED;
+      return ret;
     }
 
   return 0;
 }
 
+/* returns the 0 on success or a negative value.
+ */
+int _gnutls_negotiate_version( gnutls_session_t session, gnutls_protocol_t adv_version)
+{
+int ret;
+
+  /* if we do not support that version  */
+  if (_gnutls_version_is_supported (session, adv_version) == 0)
+    {
+      /* If he requested something we do not support
+       * then we send him the highest we support.
+       */
+      ret = _gnutls_version_max (session);
+      if (ret == GNUTLS_VERSION_UNKNOWN)
+	{
+	  /* this check is not really needed.
+	   */
+	  gnutls_assert ();
+	  return GNUTLS_E_UNKNOWN_CIPHER_SUITE;
+	}
+    }
+  else
+    {
+      ret = adv_version;
+    }
+
+  _gnutls_set_current_version (session, ret);
+    
+  return ret;
+}
+
+int _gnutls_user_hello_func( gnutls_session session, gnutls_protocol_t adv_version)
+{
+int ret;
+
+  if (session->internals.user_hello_func != NULL) 
+    {
+      ret = session->internals.user_hello_func( session);
+      if (ret < 0) 
+        {
+          gnutls_assert();
+          return ret;
+        }
+      /* Here we need to renegotiate the version since the callee might
+       * have disabled some TLS versions.
+       */
+      ret = _gnutls_negotiate_version( session, adv_version);
+      if (ret < 0) {
+        gnutls_assert();
+        return ret;
+      }
+    }
+  return 0;
+}
 
 /* Read a client hello packet. 
  * A client hello must be a known version client hello
@@ -282,13 +340,13 @@ int
 _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
 			   int datalen)
 {
-  uint8_t session_id_len, z;
+  uint8_t session_id_len;
   int pos = 0, ret;
-  uint16_t suite_size;
-  gnutls_protocol_t version;
+  uint16_t suite_size, comp_size;
+  gnutls_protocol_t adv_version;
+  int neg_version;
   int len = datalen;
-  opaque rnd[TLS_RANDOM_SIZE], *suite_ptr;
-  gnutls_protocol_t ver;
+  opaque rnd[TLS_RANDOM_SIZE], *suite_ptr, *comp_ptr;
 
   if (session->internals.v2_hello != 0)
     {				/* version 2.0 */
@@ -299,31 +357,16 @@ _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
   _gnutls_handshake_log ("HSK[%x]: Client's version: %d.%d\n", session,
 			 data[pos], data[pos + 1]);
 
-  version = _gnutls_version_get (data[pos], data[pos + 1]);
+  adv_version = _gnutls_version_get (data[pos], data[pos + 1]);
   set_adv_version (session, data[pos], data[pos + 1]);
   pos += 2;
 
-  /* if we do not support that version  */
-  if (_gnutls_version_is_supported (session, version) == 0)
+  neg_version = _gnutls_negotiate_version( session, adv_version);
+  if (neg_version < 0)
     {
-      /* If he requested something we do not support
-       * then we send him the highest we support.
-       */
-      ver = _gnutls_version_max (session);
-      if (ver == GNUTLS_VERSION_UNKNOWN)
-	{
-	  /* this check is not really needed.
-	   */
-	  gnutls_assert ();
-	  return GNUTLS_E_UNKNOWN_CIPHER_SUITE;
-	}
+      gnutls_assert();
+      return neg_version;
     }
-  else
-    {
-      ver = version;
-    }
-
-  _gnutls_set_current_version (session, ver);
 
   /* Read client random value.
    */
@@ -347,6 +390,7 @@ _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
       return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
     }
   DECR_LEN (len, session_id_len);
+  
   ret = _gnutls_server_restore_session (session, &data[pos], session_id_len);
   pos += session_id_len;
 
@@ -354,7 +398,7 @@ _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
     {				/* resumed! */
       resume_copy_required_values (session);
       session->internals.resumed = RESUME_TRUE;
-      return 0;
+      return _gnutls_user_hello_func( session, adv_version);
     }
   else
     {
@@ -376,26 +420,37 @@ _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
   suite_ptr = &data[pos];
   pos += suite_size;
 
-  /* Select an appropriate compression method
+  /* Point to the compression methods
    */
   DECR_LEN (len, 1);
-  z = data[pos++];		/* z is the number of compression methods */
+  comp_size = data[pos++];		/* z is the number of compression methods */
 
-  DECR_LEN (len, z);
-  ret = _gnutls_server_select_comp_method (session, &data[pos], z);
-  pos += z;
-
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
+  DECR_LEN (len, comp_size);
+  comp_ptr = &data[pos];
+  pos += comp_size;
 
   /* Parse the extensions (if any)
    */
-  if (ver >= GNUTLS_TLS1)
+  if (neg_version >= GNUTLS_TLS1)
     {
-      ret = _gnutls_parse_extensions (session, &data[pos], len);	/* len is the rest of the parsed length */
+      ret = _gnutls_parse_extensions (session, EXTENSION_APPLICATION, &data[pos], len);	/* len is the rest of the parsed length */
+      if (ret < 0)
+	{
+	  gnutls_assert ();
+	  return ret;
+	}
+    }
+
+  ret = _gnutls_user_hello_func( session, adv_version);
+  if (ret < 0) 
+    {
+      gnutls_assert();
+      return ret;
+    }
+  
+  if (neg_version >= GNUTLS_TLS1)
+    {
+      ret = _gnutls_parse_extensions (session, EXTENSION_TLS, &data[pos], len);	/* len is the rest of the parsed length */
       if (ret < 0)
 	{
 	  gnutls_assert ();
@@ -412,6 +467,14 @@ _gnutls_read_client_hello (gnutls_session_t session, opaque * data,
       return ret;
     }
 
+  /* select appropriate compression method */
+  ret = _gnutls_server_select_comp_method (session, comp_ptr, comp_size);
+  if (ret < 0)
+    {
+      gnutls_assert ();
+      return ret;
+    }
+
   return 0;
 }
 
@@ -424,8 +487,7 @@ _gnutls_handshake_hash_pending (gnutls_session_t session)
   int ret;
   opaque *data;
 
-  if (session->internals.handshake_mac_handle_sha == NULL ||
-      session->internals.handshake_mac_handle_md5 == NULL)
+  if (session->internals.handshake_mac_handle_init == 0)
     {
       gnutls_assert ();
       return GNUTLS_E_INTERNAL_ERROR;
@@ -441,8 +503,8 @@ _gnutls_handshake_hash_pending (gnutls_session_t session)
 
   if (siz > 0)
     {
-      _gnutls_hash (session->internals.handshake_mac_handle_sha, data, siz);
-      _gnutls_hash (session->internals.handshake_mac_handle_md5, data, siz);
+      _gnutls_hash (&session->internals.handshake_mac_handle_sha, data, siz);
+      _gnutls_hash (&session->internals.handshake_mac_handle_md5, data, siz);
     }
 
   _gnutls_handshake_buffer_empty (session);
@@ -829,9 +891,9 @@ _gnutls_handshake_hash_add_sent (gnutls_session_t session,
 
   if (type != GNUTLS_HANDSHAKE_HELLO_REQUEST)
     {
-      _gnutls_hash (session->internals.handshake_mac_handle_sha, dataptr,
+      _gnutls_hash (&session->internals.handshake_mac_handle_sha, dataptr,
 		    datalen);
-      _gnutls_hash (session->internals.handshake_mac_handle_md5, dataptr,
+      _gnutls_hash (&session->internals.handshake_mac_handle_md5, dataptr,
 		    datalen);
     }
 
@@ -872,7 +934,7 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
 
   /* first run */
   datasize = i_datasize + HANDSHAKE_HEADER_SIZE;
-  data = gnutls_alloca (datasize);
+  data = gnutls_malloc (datasize);
   if (data == NULL)
     {
       gnutls_assert ();
@@ -897,7 +959,7 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
 	 _gnutls_handshake_hash_add_sent (session, type, data, datasize)) < 0)
       {
 	gnutls_assert ();
-	gnutls_afree (data);
+	gnutls_free (data);
 	return ret;
       }
 
@@ -907,7 +969,7 @@ _gnutls_send_handshake (gnutls_session_t session, void *i_data,
     _gnutls_handshake_io_send_int (session, GNUTLS_HANDSHAKE, type,
 				   data, datasize);
 
-  gnutls_afree (data);
+  gnutls_free (data);
 
   return ret;
 }
@@ -942,6 +1004,14 @@ _gnutls_recv_handshake_header (gnutls_session_t session,
 
       *recv_type = session->internals.handshake_header_buffer.recv_type;
 
+      if (*recv_type != type)
+	{
+	  gnutls_assert ();
+	  _gnutls_handshake_log
+	    ("HSK[%x]: Handshake type mismatch (under attack?)\n", session);
+	  return GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET;
+	}
+
       return session->internals.handshake_header_buffer.packet_length;
     }
 
@@ -1115,17 +1185,6 @@ _gnutls_recv_handshake (gnutls_session_t session, uint8_t ** data,
   if (ret < 0)
     {
 
-      /* In SRP when expecting the server hello we may receive
-       * an alert instead. Do as the draft demands.
-       */
-      if (ret == GNUTLS_E_WARNING_ALERT_RECEIVED &&
-	  gnutls_alert_get (session) == GNUTLS_A_MISSING_SRP_USERNAME &&
-	  type == GNUTLS_HANDSHAKE_SERVER_HELLO)
-	{
-	  gnutls_assert ();
-	  return GNUTLS_E_INT_HANDSHAKE_AGAIN;
-	}
-
       if (ret == GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET
 	  && optional == OPTIONAL_PACKET)
 	{
@@ -1496,7 +1555,7 @@ _gnutls_read_server_hello (gnutls_session_t session,
    */
   if (version >= GNUTLS_TLS1)
     {
-      ret = _gnutls_parse_extensions (session, &data[pos], len);	/* len is the rest of the parsed length */
+      ret = _gnutls_parse_extensions (session, EXTENSION_ANY, &data[pos], len);	/* len is the rest of the parsed length */
       if (ret < 0)
 	{
 	  gnutls_assert ();
@@ -1839,19 +1898,19 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
 	  session->security_parameters.extensions.srp_username[0] == 0)
 	{
 	  /* The peer didn't send a valid SRP extension with the
-	   * SRP username. The draft requires that we send an
-	   * alert and start the handshake again.
+	   * SRP username. The draft requires that we send a fatal
+	   * alert and abort.
 	   */
 	  gnutls_assert ();
-	  ret = gnutls_alert_send (session, GNUTLS_AL_WARNING,
-				   GNUTLS_A_MISSING_SRP_USERNAME);
+	  ret = gnutls_alert_send (session, GNUTLS_AL_FATAL,
+				   GNUTLS_A_UNKNOWN_PSK_IDENTITY);
 	  if (ret < 0)
 	    {
 	      gnutls_assert ();
 	      return ret;
 	    }
 
-	  return GNUTLS_E_INT_HANDSHAKE_AGAIN;
+	  return GNUTLS_E_ILLEGAL_SRP_USERNAME;
 	}
     }
 #endif
@@ -1868,7 +1927,7 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
 	  return extdatalen;
 	}
 
-      data = gnutls_alloca (datalen + extdatalen);
+      data = gnutls_malloc (datalen + extdatalen);
       if (data == NULL)
 	{
 	  gnutls_assert ();
@@ -1916,7 +1975,7 @@ _gnutls_send_server_hello (gnutls_session_t session, int again)
   ret =
     _gnutls_send_handshake (session, data, datalen,
 			    GNUTLS_HANDSHAKE_SERVER_HELLO);
-  gnutls_afree (data);
+  gnutls_free (data);
 
   return ret;
 }
@@ -1995,25 +2054,27 @@ _gnutls_recv_hello (gnutls_session_t session, opaque * data, int datalen)
  */
 
 /**
-  * gnutls_rehandshake - This function will renegotiate security parameters
+  * gnutls_rehandshake - renegotiate security parameters
   * @session: is a #gnutls_session_t structure.
   *
   * This function will renegotiate security parameters with the
-  * client. This should only be called in case of a server.
+  * client.  This should only be called in case of a server.
   *
   * This message informs the peer that we want to renegotiate
   * parameters (perform a handshake).
   *
-  * If this function succeeds (returns 0), you must call
-  * the gnutls_handshake() function in order to negotiate
-  * the new parameters.
+  * If this function succeeds (returns 0), you must call the
+  * gnutls_handshake() function in order to negotiate the new
+  * parameters.
   *
-  * If the client does not wish to renegotiate parameters he
-  * will should with an alert message, thus the return code will be
-  * GNUTLS_E_WARNING_ALERT_RECEIVED and the alert will be
-  * GNUTLS_A_NO_RENEGOTIATION. A client may also choose to ignore
+  * If the client does not wish to renegotiate parameters he will
+  * should with an alert message, thus the return code will be
+  * %GNUTLS_E_WARNING_ALERT_RECEIVED and the alert will be
+  * %GNUTLS_A_NO_RENEGOTIATION.  A client may also choose to ignore
   * this message.
   *
+  * Returns: %GNUTLS_E_SUCCESS on success, otherwise an error.
+  *
   **/
 int
 gnutls_rehandshake (gnutls_session_t session)
@@ -2059,27 +2120,25 @@ inline static int
 _gnutls_handshake_hash_init (gnutls_session_t session)
 {
 
-  if (session->internals.handshake_mac_handle_md5 == NULL)
+  if (session->internals.handshake_mac_handle_init == 0)
     {
-      session->internals.handshake_mac_handle_md5 =
-	_gnutls_hash_init (GNUTLS_MAC_MD5);
+      int ret =
+	_gnutls_hash_init (&session->internals.handshake_mac_handle_md5, GNUTLS_MAC_MD5);
 
-      if (session->internals.handshake_mac_handle_md5 == GNUTLS_HASH_FAILED)
+      if (ret < 0)
 	{
 	  gnutls_assert ();
-	  return GNUTLS_E_MEMORY_ERROR;
+	  return ret;
 	}
-    }
 
-  if (session->internals.handshake_mac_handle_sha == NULL)
-    {
-      session->internals.handshake_mac_handle_sha =
-	_gnutls_hash_init (GNUTLS_MAC_SHA1);
-      if (session->internals.handshake_mac_handle_sha == GNUTLS_HASH_FAILED)
+      ret = _gnutls_hash_init(&session->internals.handshake_mac_handle_sha, GNUTLS_MAC_SHA1);
+      if (ret < 0)
 	{
 	  gnutls_assert ();
 	  return GNUTLS_E_MEMORY_ERROR;
 	}
+	
+      session->internals.handshake_mac_handle_init = 1;
     }
 
   return 0;
@@ -2149,24 +2208,27 @@ _gnutls_recv_supplemental (gnutls_session_t session)
   * gnutls_handshake - This is the main function in the handshake protocol.
   * @session: is a #gnutls_session_t structure.
   *
-  * This function does the handshake of the TLS/SSL protocol,
-  * and initializes the TLS connection. 
+  * This function does the handshake of the TLS/SSL protocol, and
+  * initializes the TLS connection.
   *
-  * This function will fail if any problem is encountered,
-  * and will return a negative error code. In case of a client,
-  * if the client has asked to resume a session, but the server couldn't, 
-  * then a full handshake will be performed.
+  * This function will fail if any problem is encountered, and will
+  * return a negative error code. In case of a client, if the client
+  * has asked to resume a session, but the server couldn't, then a
+  * full handshake will be performed.
   *
-  * The non-fatal errors such as GNUTLS_E_AGAIN and GNUTLS_E_INTERRUPTED
-  * interrupt the handshake procedure, which should be later be resumed.
-  * Call this function again, until it returns 0; cf.
-  * gnutls_record_get_direction() and gnutls_error_is_fatal().
+  * The non-fatal errors such as %GNUTLS_E_AGAIN and
+  * %GNUTLS_E_INTERRUPTED interrupt the handshake procedure, which
+  * should be later be resumed.  Call this function again, until it
+  * returns 0; cf.  gnutls_record_get_direction() and
+  * gnutls_error_is_fatal().
   *
-  * If this function is called by a server after a rehandshake request then
-  * GNUTLS_E_GOT_APPLICATION_DATA or GNUTLS_E_WARNING_ALERT_RECEIVED 
-  * may be returned. Note that these are non fatal errors, only in the
-  * specific case of a rehandshake. Their meaning is that the client
-  * rejected the rehandshake request.
+  * If this function is called by a server after a rehandshake request
+  * then %GNUTLS_E_GOT_APPLICATION_DATA or
+  * %GNUTLS_E_WARNING_ALERT_RECEIVED may be returned.  Note that these
+  * are non fatal errors, only in the specific case of a rehandshake.
+  * Their meaning is that the client rejected the rehandshake request.
+  *
+  * Returns: %GNUTLS_E_SUCCESS on success, otherwise an error.
   *
   **/
 int
@@ -2217,21 +2279,8 @@ gnutls_handshake (gnutls_session_t session)
   return 0;
 }
 
-/* Here if GNUTLS_E_INT_HANDSHAKE_AGAIN is received we go to
- * restart. This works because this error code may only be
- * received on the first 2 handshake packets. If for some reason
- * this changes we should return GNUTLS_E_AGAIN.
- */
 #define IMED_RET( str, ret) do { \
 	if (ret < 0) { \
-		if (ret == GNUTLS_E_INT_HANDSHAKE_AGAIN && \
-			session->internals.handshake_restarted == 1) \
-			ret = GNUTLS_E_INTERNAL_ERROR; \
-		if (ret == GNUTLS_E_INT_HANDSHAKE_AGAIN) { \
-			STATE = STATE0; \
-			session->internals.handshake_restarted = 1; \
-			goto restart; \
-		} \
 		if (gnutls_error_is_fatal(ret)==0) return ret; \
 		gnutls_assert(); \
 		ERR( str, ret); \
@@ -2263,7 +2312,6 @@ _gnutls_handshake_client (gnutls_session_t session)
 					    session_id_size, buf,
 					    sizeof (buf)));
 #endif
-restart:
 
   switch (STATE)
     {
@@ -2491,8 +2539,6 @@ _gnutls_handshake_server (gnutls_session_t session)
 {
   int ret = 0;
 
-restart:
-
   switch (STATE)
     {
     case STATE0:
@@ -2594,8 +2640,6 @@ _gnutls_handshake_common (gnutls_session_t session)
 {
   int ret = 0;
 
-restart:
-
   /* send and recv the change cipher spec and finished messages */
   if ((session->internals.resumed == RESUME_TRUE
        && session->security_parameters.entity == GNUTLS_CLIENT)
@@ -2636,11 +2680,13 @@ int
 _gnutls_generate_session_id (opaque * session_id, uint8_t * len)
 {
   *len = TLS_MAX_SESSION_ID_SIZE;
+  int ret;
 
-  if (gc_nonce (session_id, *len) != GC_OK)
+  ret = _gnutls_rnd (RND_NONCE, session_id, *len);
+  if (ret < 0)
     {
       gnutls_assert ();
-      return GNUTLS_E_RANDOM_FAILED;
+      return ret;
     }
 
   return 0;
@@ -2836,7 +2882,7 @@ _gnutls_remove_unwanted_ciphersuites (gnutls_session_t session,
   /* get all the key exchange algorithms that are 
    * supported by the X509 certificate parameters.
    */
-  if (cert_cred != NULL && (ret =
+  if ((ret =
        _gnutls_selected_cert_supported_kx (session, &alg, &alg_size)) < 0)
     {
       gnutls_assert ();
@@ -2918,16 +2964,19 @@ _gnutls_remove_unwanted_ciphersuites (gnutls_session_t session,
 }
 
 /**
-  * gnutls_handshake_set_max_packet_length - This function will set the maximum length of a handshake message
-  * @session: is a #gnutls_session_t structure.
-  * @max: is the maximum number.
-  *
-  * This function will set the maximum size of a handshake message.
-  * Handshake messages over this size are rejected.
-  * The default value is 16kb which is large enough. Set this to 0 if you do not want
-  * to set an upper limit.
-  *
-  **/
+ * gnutls_handshake_set_max_packet_length - set the maximum size of the handshake
+ * @session: is a #gnutls_session_t structure.
+ * @max: is the maximum number.
+ *
+ * This function will set the maximum size of all handshake messages.
+ * Handshakes over this size are rejected with
+ * %GNUTLS_E_HANDSHAKE_TOO_LARGE error code.  The default value is
+ * 48kb which is typically large enough.  Set this to 0 if you do not
+ * want to set an upper limit.
+ *
+ * The reason for restricting the handshake message sizes are to
+ * limit Denial of Service attacks.
+ **/
 void
 gnutls_handshake_set_max_packet_length (gnutls_session_t session, size_t max)
 {
@@ -2952,11 +3001,15 @@ _gnutls_get_adv_version (gnutls_session_t session)
   * gnutls_handshake_get_last_in - Returns the last handshake message received.
   * @session: is a #gnutls_session_t structure.
   *
-  * Returns the last handshake message received. This function is only useful
-  * to check where the last performed handshake failed. If the previous handshake
-  * succeed or was not performed at all then no meaningful value will be returned.
+  * This function is only useful to check where the last performed
+  * handshake failed.  If the previous handshake succeed or was not
+  * performed at all then no meaningful value will be returned.
   *
-  * Check gnutls.h for the available handshake descriptions.
+  * Check %gnutls_handshake_description_t in gnutls.h for the
+  * available handshake descriptions.
+  *
+  * Returns: the last handshake message type received, a
+  * %gnutls_handshake_description_t.
   **/
 gnutls_handshake_description_t
 gnutls_handshake_get_last_in (gnutls_session_t session)
@@ -2968,12 +3021,15 @@ gnutls_handshake_get_last_in (gnutls_session_t session)
   * gnutls_handshake_get_last_out - Returns the last handshake message sent.
   * @session: is a #gnutls_session_t structure.
   *
-  * Returns the last handshake message sent. This function is only useful
-  * to check where the last performed handshake failed. If the previous handshake
-  * succeed or was not performed at all then no meaningful value will be returned.
+  * This function is only useful to check where the last performed
+  * handshake failed.  If the previous handshake succeed or was not
+  * performed at all then no meaningful value will be returned.
   *
-  * Check gnutls.h for the available handshake descriptions.
+  * Check %gnutls_handshake_description_t in gnutls.h for the
+  * available handshake descriptions.
   *
+  * Returns: the last handshake message type sent, a
+  * %gnutls_handshake_description_t.
   **/
 gnutls_handshake_description_t
 gnutls_handshake_get_last_out (gnutls_session_t session)
