@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005 Free Software Foundation
+ * Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2008 Free Software Foundation
  *
- * Author: Nikos Mavroyanopoulos
+ * Author: Nikos Mavrogiannopoulos
  *
  * This file is part of GNUTLS.
  *
@@ -59,10 +59,16 @@ int _gnutls_rsa_export_set_pubkey (gnutls_session_t session,
 int _gnutls_session_is_resumable (gnutls_session_t session);
 int _gnutls_session_is_export (gnutls_session_t session);
 
+int _gnutls_session_is_psk (gnutls_session_t session);
+
 int _gnutls_openpgp_send_fingerprint (gnutls_session_t session);
 
-int _gnutls_PRF (const opaque * secret, int secret_size, const char *label,
-		 int label_size, const opaque * seed, int seed_size,
+int _gnutls_PRF (gnutls_session_t session,
+		 const opaque * secret, int secret_size,
+		 const char *label, int label_size,
+		 const opaque * seed, int seed_size,
 		 int total_bytes, void *ret);
 
+int gnutls_init (gnutls_session_t * session, gnutls_connection_end_t con_end);
+
 #define DEFAULT_CERT_TYPE GNUTLS_CRT_X509
